package application;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.font.*;
import java.awt.geom.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.undo.*;
import javax.swing.filechooser.*;

import project.*;
import language.*;
import static logging.Logging.logger;

public class EditorPane extends JEditorPane implements Observer {

    private String path = null;
    private Font font = null;
    private UndoManager undoManager;
    private boolean saved = true;
    public static final int LIMIT = 1024;

    JPopupMenu popup = null;
    PopupAction action = null;

    public EditorPane() {

        setBackground(Color.white);
        font = new Font("Monospaced", Font.PLAIN, 12);
        setFont(font);
        setEditable(true);

        addKeyListener(new KeyAction());
        addFocusListener(new FocusAction());
        addMouseListener(new MouseAction());

        popup = new JPopupMenu();
        action = new PopupAction();

        JMenuItem item;

        item = new JMenuItem("Undo");
        item.addActionListener(action);
        popup.add(item);
        item = new JMenuItem("Redo");
        item.addActionListener(action);
        popup.add(item);
        popup.addSeparator();

        item = new JMenuItem("Cut");
        item.addActionListener(action);
        popup.add(item);
        item = new JMenuItem("Copy");
        item.addActionListener(action);
        popup.add(item);
        item = new JMenuItem("Paste");
        item.addActionListener(action);
        popup.add(item);

        add(popup);

    }

    public boolean close() {
        if (!saved) {
            if (Config.AUTOSAVE)
                save();
            else {
                int option = JOptionPane.showConfirmDialog(this, path + " has been modified. Save changes?", "Save", JOptionPane.YES_NO_CANCEL_OPTION);
                switch (option) {
                    case JOptionPane.NO_OPTION :
                        break;
                    case JOptionPane.YES_OPTION :
                        save();
                        break;
                    default : return false;
                }
            }
        }

        ProjectManager.closeFile(path);
        return true;
    }

    public void save() {
        ProjectManager.saveFile(path, getText());
        setSaved();
    }

    public void undo() {
        try {
            undoManager.undo();
        } catch (CannotUndoException e) {
        }
    }

    public void redo() {
        try {
            undoManager.redo();
        } catch (CannotRedoException e) {
        }
    }

    public void compile() {
        ProjectManager.compileProjectForFile(path);
    }

    public void run() {
        run(null);
    }

    public void run(String args) {

        if (! ProjectManager.compileAndRunProjectForFile(path, args)) {

            JFileChooser chooser = new JFileChooser();
            FileFilter filter = new FileFilter() {
                public boolean accept(java.io.File f) {
                    return f.getName().toLowerCase().endsWith(".class") || f.isDirectory();
                }
                public String getDescription() {
                    return "Class Files";
                }
            };
            chooser.setFileFilter(filter);
            chooser.setCurrentDirectory(new java.io.File(ProjectManager.getProjectBinPathForFile(path)));

            if (chooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
                ProjectManager.runProjectMain(chooser.getSelectedFile().getAbsolutePath(), args);
            } else {
                return;
            }
        }

    }

    public void goToLine() {
        new GotoLineDialog();
    }

    public void setSaved() {
        if (!saved) {
            EditorContent editorContent = (EditorContent) getParent().getParent();
            Editor editor = (Editor) editorContent.getParent();
            ((EditorTab) editor.getTabComponentAt(editor.indexOfComponent(editorContent))).setSaved();
            saved = true;
        }
    }

    public String getPath() {
        return path;
    }

    public void update(Observable obs, Object args) {
        File file = (File) obs;
        path = file.getPath();
        setContentType(file.getContentType());
        setText(file.getContent());

        undoManager = new UndoManager();
        undoManager.setLimit(LIMIT);
        getDocument().addUndoableEditListener(undoManager);
        getDocument().addDocumentListener(new DocumentAction());

        setCaretPosition(0);
        grabFocus();
    }

    public void submitData() {
        String data = CCPopup.getSubmitData();
        try {
            if (data.length() > 0 && data.charAt(data.length() - 1) == '(') {
                getDocument().insertString(getCaretPosition(),
                        data.substring(0, data.length() - 1), null);
                getDocument().insertString(getCaretPosition(), "(", null);
            } else {
                getDocument().insertString(getCaretPosition(), data, null);
            }
        } catch (BadLocationException e) {
            logger.throwing("application.EditorPane", "submitData", e);
            return;
        }
    }

    public class KeyAction implements KeyListener {

        public void keyPressed(KeyEvent e) {
            EditorPane editorPane = (EditorPane) e.getSource();

            switch (e.getKeyCode()) {

                case KeyEvent.VK_ENTER:
                    if (CCPopup.isVisible()) {
                        EditorPane editor = (EditorPane) e.getSource();
                        editor.submitData();
                        CCPopup.hide();
                        e.consume();
                    }
                    break;

                case KeyEvent.VK_LEFT:
                    if(e.isControlDown()) {
                        EditorContent editorContent = (EditorContent) editorPane.getParent().getParent();
                        Editor editor = (Editor) editorContent.getParent();

                        int index = editor.indexOfComponent(editorContent);
                        if (index == 0)
                            index = editor.getTabCount();

                        editor.setSelectedIndex(index - 1);
                     }
                    CCPopup.hide();
                    break;

                case KeyEvent.VK_RIGHT:
                    if(e.isControlDown()) {
                        EditorContent editorContent = (EditorContent) editorPane.getParent().getParent();
                        Editor editor = (Editor) editorContent.getParent();

                        int index = editor.indexOfComponent(editorContent);
                        if (index == editor.getTabCount() - 1)
                            index = -1;

                        editor.setSelectedIndex(index + 1);
                     }
                    CCPopup.hide();
                    break;

                case KeyEvent.VK_PAGE_DOWN:
                case KeyEvent.VK_PAGE_UP:
                case KeyEvent.VK_ESCAPE:
                    CCPopup.hide();
                    break;

                case KeyEvent.VK_DOWN:
                    if (CCPopup.isVisible()) {
                        CCPopup.selectNext();
                        e.consume();
                    }
                    break;

                case KeyEvent.VK_UP:
                    if (CCPopup.isVisible()) {
                        CCPopup.selectPrev();
                        e.consume();
                    }
                    break;

                case KeyEvent.VK_PERIOD:
                    Point p = editorPane.getCaret().getMagicCaretPosition();
                    SwingUtilities.convertPointToScreen(p, editorPane);
                    CCPopup.setPosition(p.x, p.y + 15);
                    break;

                case KeyEvent.VK_SPACE:
                    if (e.isControlDown()) {
                        int filterSize = ((XDocument) getDocument()).startParserCurrentScope(getCaretPosition());
                        Rectangle2D rect = font.getStringBounds("X", new FontRenderContext(null, false, false));
                        p = editorPane.getCaret().getMagicCaretPosition();
                        SwingUtilities.convertPointToScreen(p, editorPane);
                        CCPopup.setPosition((int) (p.x - filterSize * rect.getWidth()), p.y + 15);
                    }
                    break;
            }
        }

        public void keyReleased(KeyEvent e) {
        }

        public void keyTyped(KeyEvent e) {
        }

    }

    public class FocusAction implements FocusListener {

        public void focusGained(FocusEvent e) {
        }

        public void focusLost(FocusEvent e) {
            CCPopup.hide();
        }

    }

    public class MouseAction implements MouseListener {

        public void mouseClicked(MouseEvent e) {
        }

        public void mouseEntered(MouseEvent e) {
        }

        public void mouseExited(MouseEvent e) {
        }

        public void mousePressed(MouseEvent e) {
            CCPopup.hide();
        }

        public void mouseReleased(MouseEvent e) {
        }

    }

    public void processMouseEvent(MouseEvent e) {
        super.processMouseEvent(e);
        if(e.isPopupTrigger()) {
            popup.show(e.getComponent(), e.getX(), e.getY());
        }
    }

    public class DocumentAction implements DocumentListener {

        private void setUnsaved() {
            if (saved) {
                EditorContent editorContent = (EditorContent) getParent().getParent();
                Editor editor = (Editor) editorContent.getParent();
                ((EditorTab) editor.getTabComponentAt(editor.indexOfComponent(editorContent))).setUnsaved();
                saved = false;
            }
        }

        public void changedUpdate(DocumentEvent arg0) {
            setUnsaved();
        }

        public void insertUpdate(DocumentEvent arg0) {
            setUnsaved();
        }

        public void removeUpdate(DocumentEvent arg0) {
            setUnsaved();
        }

    }

    public class PopupAction implements ActionListener {

        public void actionPerformed(ActionEvent e) {

            String command = e.getActionCommand();

            if (command.equals("Cut"))
                cut();
            else if (command.equals("Copy"))
                copy();
            else if (command.equals("Paste"))
                paste();
            else if (command.equals("Undo"))
                undo();
            else if(command.equals("Redo"))
                redo();
        }

    }

    public class GotoLineDialog extends JDialog implements ActionListener, KeyListener {
        private int lines;
        private JTextField text;

        public GotoLineDialog() {
            super(Application.getFrame(), "Goto Line");

            JPanel topPanel = new JPanel();
            PlainDocument doc = (PlainDocument) getDocument();
            Element root = doc.getDefaultRootElement();

            text = new JTextField((new Integer(root.getElementIndex(getCaretPosition()) + 1)).toString());
            text.selectAll();
            text.addKeyListener(this);
            lines = root.getElementCount();

            topPanel.setLayout(new GridLayout(2, 1));
            topPanel.add(new JLabel("Enter line number (1.." + lines + ") :"));
            topPanel.add(text);

            JPanel buttonsPanel = new JPanel();
            JButton ok = new JButton("Ok");
            ok.addActionListener(this);
            JButton cancel = new JButton("Cancel");
            cancel.addActionListener(this);

            buttonsPanel.add(ok);
            buttonsPanel.add(cancel);

            setLayout(new BorderLayout());
            add(topPanel, BorderLayout.NORTH);
            add(buttonsPanel, BorderLayout.SOUTH);

            text.grabFocus();

            addKeyListener(this);
            setLocationRelativeTo(null);
            setVisible(true);
            pack();
        }

        public void actionPerformed(ActionEvent e) {
            if (e.getActionCommand() == "Ok")
                try {
                    int value = Integer.parseInt(text.getText());
                    if (value > 0 && value <= lines) {
                        Element root = getDocument().getDefaultRootElement();
                        setCaretPosition(root.getElement(value - 1).getStartOffset());
                    }
                } catch (NumberFormatException ex) {
                    dispose();
                }
            dispose();
        }

        public void keyPressed(KeyEvent e) {
            if (e.getKeyCode() == KeyEvent.VK_ENTER) {
                try {
                    int value = Integer.parseInt(text.getText());
                    if (value > 0 && value <= lines) {
                        Element root = getDocument().getDefaultRootElement();
                        setCaretPosition(root.getElement(value - 1).getStartOffset());
                    }
                    dispose();
                } catch (NumberFormatException ex) {
                    dispose();
                }
            } else
                if (e.getKeyCode() == KeyEvent.VK_ESCAPE)
                    dispose();
        }

        public void keyReleased(KeyEvent e) {
        }

        public void keyTyped(KeyEvent e) {
        }

    }

}

