package language.java;

import java.awt.*;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentEvent.ElementChange;
import javax.swing.text.*;

import application.Config;
import static logging.Logging.logger;

import language.java.JavaDocument.CommentAttribute;

public class JavaContext extends StyleContext implements ViewFactory {

    JavaColorScheme colors;

    public JavaContext() {
        super();
        updateColorScheme();
    }

    public JavaColorScheme getJavaColorScheme() {
        return colors;
    }

    public void updateColorScheme() {
        // see if we have a ColorScheme set
        if (Config.javaColorScheme != null && Config.javaColorScheme.equalsIgnoreCase("JavaShineColorScheme")) {
            colors = new JavaShineColorScheme();
            return;
        }
        // if we don't use the default
        colors = new JavaDefaultColorScheme();
    }

    public View create(Element elem) {
        return new JavaView(elem);
    }

    class JavaView extends WrappedPlainView {

        JavaDocument.Scanner scanner = new JavaDocument.Scanner();

        public JavaView(Element elem) {
            super(elem, true);
        }

        public void paint(Graphics g, Shape a) {
            super.paint(g, a);
            scanner.setValid(false);
        }

        public void insertUpdate(DocumentEvent chng, Shape a, ViewFactory f) {
            super.insertUpdate(chng, a, f);

            // updates elements comment attributes
            JavaDocument doc = (JavaDocument) getDocument();
            Element root = doc.getDefaultRootElement();
            ElementChange changes = chng.getChange(root);

            if (changes != null) {
                Element[] elems = changes.getChildrenAdded();
                doc.updateComments(elems);
            } else {
                Element[] elems = new Element[1];
                elems[0] = root.getElement(root.getElementIndex(chng.getOffset()));
                doc.updateComments(elems);
            }

            getContainer().repaint();
        }

        public void removeUpdate(DocumentEvent chng, Shape a, ViewFactory f) {
            super.removeUpdate(chng, a, f);

            // updates elements comment attributes
            JavaDocument doc = (JavaDocument) getDocument();
            Element root = doc.getDefaultRootElement();
            Element[] elems = new Element[1];
            elems[0] = root.getElement(root.getElementIndex(chng.getOffset()));
            doc.updateComments(elems);

            getContainer().repaint();
        }

        protected int drawUnselectedText(Graphics g, int x, int y, int startPos, int endPos) throws BadLocationException {
            return draw(g, x, y, startPos, endPos, true);
        }

        protected int drawSelectedText(Graphics g, int x, int y, int startPos, int endPos) throws BadLocationException {
            return draw(g, x, y, startPos, endPos, false);
        }

        private int draw(Graphics g, int x, int y, int startPos, int endPos, boolean withColors)
                                                            throws BadLocationException {
            Document doc = getDocument();

            Font font = g.getFont();
            if (!withColors)
                g.setColor(Color.WHITE);

            Color lastColor = null;
            int lastStyle = 0;
            int lastPos = startPos;

            while (startPos < endPos) {
                scanner.restartScan(doc, startPos);

                Color actColor = colors.getColor(scanner.token);
                int actStyle = colors.getStyle(scanner.token);

                int tokenStart = (int) scanner.getStartPos();
                int tokenEnd = (int) scanner.getEndPos();

                String token = null;
                try {
                    token = doc.getText(tokenStart, tokenEnd - tokenStart);
                } catch (Exception e) {
                    logger.throwing("JavaView", "draw", e);
                }

                // see if we are in a multi-line comment

                Element elem = doc.getDefaultRootElement();
                elem = elem.getElement(elem.getElementIndex((int) scanner.getStartPos()));
                if (elem != null) {
                    AttributeSet attributes = elem.getAttributes();

                    if (attributes.isDefined(CommentAttribute.comment) &&
                            ((CommentAttribute) attributes.getAttribute(CommentAttribute.comment)).
                                                isComment(tokenStart - elem.getStartOffset())) {
                        actColor = colors.getColor(JavaColorScheme.COMMENT);
                        actStyle = colors.getStyle(JavaColorScheme.COMMENT);
                    }
                }

                // see if it's a single-line /*...*/ comment

                if (token.indexOf("/*") == 0 && token.indexOf("*/") == token.length() - 2) {
                    actColor = colors.getColor(JavaColorScheme.COMMENT);
                    actStyle = colors.getStyle(JavaColorScheme.COMMENT);
                }

                // if we change style we draw the buffered text
                if (lastColor != null && (actColor != lastColor || actStyle != lastStyle)) {
                    if (withColors) g.setColor(lastColor);
                    g.setFont(new Font(font.getFamily(), lastStyle, font.getSize()));
                    Segment txt = getLineBuffer();
                    doc.getText(lastPos, startPos - lastPos, txt);
                    x = Utilities.drawTabbedText(txt, x, y, g, this, lastPos);
                    lastPos = startPos;
                }

                lastColor = actColor;
                lastStyle = actStyle;
                if (startPos >= tokenEnd) startPos = endPos;
                else startPos = tokenEnd;
            }

            if (withColors) g.setColor(lastColor);
            g.setFont(new Font(font.getFamily(), lastStyle, font.getSize()));
            Segment txt = getLineBuffer();
            doc.getText(lastPos, endPos - lastPos, txt);
            x = Utilities.drawTabbedText(txt, x, y, g, this, lastPos);

            return x;
        }

    }

}
