package language.java;

import java.io.*;
import java.util.*;
import javax.swing.text.*;

import com.sun.tools.javac.main.JavaCompiler;
import com.sun.tools.javac.util.Context;
import com.sun.tools.javac.tree.JCTree;
import static com.sun.tools.javac.tree.JCTree.*;

import application.*;
import language.*;
import logging.Logging;
import static logging.Logging.logger;

public class JavaDocument extends XDocument {

    String filePath = null;

    int expectedOffset = 0;
    int state = STATE_INVALID;
    LinkedList <Character> filter = null;

    static final int STATE_INVALID = 0;
    static final int STATE_POINT_INSERTED = 1;
    static final int STATE_WRITING = 2;
    static final int STATE_FIRST_BRACKET = 3;

    public JavaDocument(String path) {
        super(new GapContent(1024));
        filePath = path;
        setDocumentFilter(new JavaDocumentFilter());
    }

    private void insertUpdateState(char ch) {
        switch (state) {

            case STATE_INVALID:
                if (ch == '.') {
                    state = STATE_POINT_INSERTED;
                    filter = new LinkedList <Character> ();
                }
                break;

            case STATE_POINT_INSERTED:
                if (Character.isJavaIdentifierStart(ch)) {
                    state = STATE_WRITING;
                    filter.add(ch);
                } else if (ch == '.') {
                    CCPopup.hide();
                } else {
                    state = STATE_INVALID;
                    filter = null;
                }
                break;

            case STATE_WRITING:
                if (Character.isJavaIdentifierPart(ch)) {
                    state = STATE_WRITING;
                    filter.add(ch);
                } else if (ch == '.') {
                    state = STATE_POINT_INSERTED;
                    filter = new LinkedList <Character> ();
                } else if (ch == '(') {
                    state = STATE_FIRST_BRACKET;
                    filter.add('(');
                } else {
                    state = STATE_INVALID;
                    filter = null;
                }
                break;

            // NOTE: When the user requests suggestions and he is writing
            // the arguments for a function, a new request will cancel the
            // old one.
            case STATE_FIRST_BRACKET:
            default:
                switch (ch) {
                    case '.':
                        state = STATE_POINT_INSERTED;
                        filter = new LinkedList <Character> ();
                        break;
                    case '(':
                        ++state;
                        break;
                    case ')':
                        if (state == STATE_FIRST_BRACKET) {
                            state = STATE_INVALID;
                            filter = null;
                        } else
                            --state;
                        break;
                    case ',':
                        if (state == STATE_FIRST_BRACKET)
                            filter.add(',');
                        break;
                }
                break;

        }
    }

    private void removeUpdateState(char ch) {
        switch (state) {

            case STATE_INVALID:
                break;

            case STATE_POINT_INSERTED:
                state = STATE_INVALID;
                filter = null;
                break;

            case STATE_WRITING:
                filter.removeLast();
                if (filter.isEmpty())
                    state = STATE_POINT_INSERTED;
                break;

            case STATE_FIRST_BRACKET:
                switch (ch) {
                    case '(':
                        filter.removeLast();
                        state = STATE_WRITING;
                        break;
                    case ')':
                        ++state;
                        break;
                    case ',':
                        filter.removeLast();
                        break;
                }
                break;

            default:
                if (ch == '(')
                    --state;
                else if (ch == ')')
                    ++state;
                break;

        }
    }

    protected synchronized void insertUpdate(DefaultDocumentEvent chng, AttributeSet attr) {

        super.insertUpdate(chng, attr);

        // Decide if we have to show the popup. This must be done before setting the filter.
        try {
            if (chng.getLength() == 1 && getText(chng.getOffset(), 1).equals("."))
                CCPopup.show();
        } catch (BadLocationException e) {
            logger.throwing("language.java.JavaDocument", "insertUpdate", e);
        }

        // Adjust the filter for code completion.
        if (expectedOffset != chng.getOffset()) {
            state = STATE_INVALID;
            filter = null;
        }

        if (chng.getLength() == 1) {
            try {
                insertUpdateState(getText(chng.getOffset(), 1).charAt(0));
            } catch (BadLocationException e) {
                logger.throwing("language.java.JavaDocument", "insertUpdate", e);
            }
        }
        else {
            state = STATE_INVALID;
            filter = null;
        }

        expectedOffset = chng.getOffset() + chng.getLength();
        if (filter != null)
            CCPopup.setFilter(filter);
        else
            CCPopup.hide();

        // Run the parser.
        try {
            if (chng.getLength() == 1 && getText(chng.getOffset(), 1).equals(".")) {
                if (parser != null && parser.isAlive())
                    parser.interrupt();
                CCPopup.show();
                parser = new JavaParser(filePath, chng.getOffset(), getText(0, getLength()), false);
                parser.start();
            }
        } catch (BadLocationException e) {
            logger.throwing("language.java.JavaDocument", "insertUpdate", e);
        }

        // Auto-complete brackets
        try {
            if (chng.getLength() == 1 && getText(chng.getOffset(), 1).equals("(") && !shouldIgnorePosition(chng.getOffset())) {
                insertString(chng.getOffset() + 1, ")", null);
            }
        } catch (BadLocationException e) {
            logger.throwing("language.java.JavaDocument", "insertUpdate", e);
        }

    }

    protected synchronized void removeUpdate(DefaultDocumentEvent chng) {

        super.removeUpdate(chng);

        // Adjust the filter for code completion.
        if (expectedOffset - 1 != chng.getOffset()) {
            state = STATE_INVALID;
            filter = null;
        }

        if (chng.getLength() == 1) {
            try {
                removeUpdateState(getText(chng.getOffset(), 1).charAt(0));
            } catch (BadLocationException e) {
                logger.throwing("language.java.JavaDocument", "removeUpdate", e);
            }
        }
        else {
            state = STATE_INVALID;
            filter = null;
        }

        expectedOffset = chng.getOffset();
        if (filter != null)
            CCPopup.setFilter(filter);
        else
            CCPopup.hide();

    }

    public synchronized int startParserCurrentScope(int caretPos) {

        int pos = caretPos - 1;

        CCPopup.hide();

        try {
            String data = getText(0, getLength());
            expectedOffset = caretPos;
            filter = new LinkedList <Character> ();

            while (pos >= 0 && Character.isJavaIdentifierPart(data.charAt(pos))) {
                filter.addFirst(data.charAt(pos));
                --pos;
            }

            while (pos >= 0 && Character.isSpaceChar(data.charAt(pos)))
                --pos;

            if (filter.size() == 0)
                state = STATE_POINT_INSERTED;
            else
                state = STATE_WRITING;

            if (pos >= 0) {

                if (parser != null && parser.isAlive())
                    parser.interrupt();

                CCPopup.show();
                CCPopup.setFilter(filter);

                if (data.charAt(pos) == '.') {
                    parser = new JavaParser(filePath, pos, data, false);
                } else {
                    parser = new JavaParser(filePath, caretPos, data, true);
                }

                parser.start();

            }

        } catch (BadLocationException e) {
            logger.throwing("language.java.JavaDocument", "startParserCurrentScope", e);
            return 0;
        }

        return filter.size();

    }

    private void buildTree(FunctionsTree tree, JCTree jctree) {
        List <JCTree> defs = null;

        if (jctree.tag == TOPLEVEL)
            defs = ((JCCompilationUnit) jctree).defs;
        if (jctree.tag == CLASSDEF)
            defs = ((JCClassDecl) jctree).defs;

        if (defs != null) {
            for (JCTree t : defs)
                if (t.tag == CLASSDEF) {
                    JCClassDecl cls = (JCClassDecl) t;
                    StringBuffer buffer = new StringBuffer();

                    buffer.append("class " + cls.getSimpleName());

                    FunctionsTree entry = new FunctionsTree(t.getStartPosition(), buffer.toString());
                    tree.list.add(entry);
                    buildTree(entry, t);
                } else if (t.tag == METHODDEF) {
                    JCMethodDecl mth = (JCMethodDecl) t;
                    StringBuffer buffer = new StringBuffer();

                    if (mth.restype != null)
                        buffer.append(mth.restype.toString());
                    buffer.append(" " + mth.name + "(");
                    for (int i = 0; i < mth.params.size(); ++i) {
                        if (i != 0)
                            buffer.append(", ");
                        buffer.append(mth.params.get(i).toString());
                    }
                    buffer.append(")");

                    FunctionsTree entry = new FunctionsTree(t.getStartPosition(), buffer.toString());
                    tree.list.add(entry);
                    buildTree(entry, t);
                }
        }
    }

    public FunctionsTree getFunctionsTree() {

        try {
            // Create a temporary file.
            File tmp = File.createTempFile(
                    "java-document-" + System.currentTimeMillis(),
                    ".java");

            // Write data to that file.
            FileOutputStream out = new FileOutputStream(tmp);
            out.write(getText(0, getLength()).getBytes());
            out.close();

            // Redirect System.out and System.err to the logger.
            Logging.startLoggingSystemOut("language.java.JavaDocument", "getFunctionsTree");
            Logging.startLoggingSystemErr("language.java.JavaDocument", "getFunctionsTree");

            // Parse the file for getting a tree.
            Context context = new Context();
            JavaCompiler compiler = new JavaCompiler(context);
            JCCompilationUnit jctree = compiler.parse(tmp.getAbsolutePath()); // FIXME: deprecated

            // Restore System.out and System.err.
            Logging.stopLoggingSystemOut();
            Logging.stopLoggingSystemErr();

            // Delete the temporary file.
            tmp.delete();

            XDocument.FunctionsTree tree = new XDocument.FunctionsTree(0, null);
            buildTree(tree, jctree);

            return tree;

        }  catch (Exception e) {
            logger.throwing("language.java.JavaDocument", "getFunctionsTree", e);
            return null;
        }

    }

    private boolean shouldIgnorePosition(int pos) {
        // see if pos is in a multi-line comment

        Element root = getDefaultRootElement();
        Element elem = root.getElement(root.getElementIndex(pos));

        pos = pos - elem.getStartOffset();

        AttributeSet attributes = elem.getAttributes();
        if (attributes != null && attributes.isDefined(CommentAttribute.comment) &&
                ((CommentAttribute) attributes.getAttribute(CommentAttribute.comment)).isComment(pos))
            return true;

        // see if pos is in a single line comment or string value
        String line = null;
        int commentStart, lineCommentStart, stringStart;
        try {
            line = getText(elem.getStartOffset(), elem.getEndOffset() - elem.getStartOffset());
        } catch (BadLocationException e) {
            logger.throwing("language.java.JavaDocument", "inComments", e);
            return false;
        }

        while (line.length() > 0) {
            if ((commentStart = line.indexOf("/*")) == -1) commentStart = line.length();
            if ((lineCommentStart = line.indexOf("//")) == -1) lineCommentStart = line.length();
            if ((stringStart = line.indexOf("\"")) == -1) stringStart = line.length();

            int min = Math.min(commentStart, Math.min(lineCommentStart, stringStart));

            if (pos < min) return false;

            if (min == lineCommentStart) return true;

            if (min == stringStart) {
                int stringEnd = line.substring(stringStart + 1).indexOf("\"");
                if (stringEnd == -1) return true;

                stringEnd += stringStart + 1;
                if (pos < stringEnd) return true;

                line = line.substring(stringEnd + 1);
                pos -= stringEnd + 1;
            }

            if (min == commentStart) {
                int commentEnd = line.substring(commentStart + 2).indexOf("*/");
                if (commentEnd == -1) return true;

                commentEnd += commentStart + 2;
                if (pos < commentEnd) return true;

                line = line.substring(commentEnd + 2);
                pos -= commentEnd + 2;
            }
        }

        return false;
    }

    private int getCommentStart(String line) {
        int commentStart = line.indexOf("/*"),
            lineCommentStart = line.indexOf("//"),
            stringStart = line.indexOf("\"");

        if (commentStart == -1) return -1;
        if (lineCommentStart == -1) lineCommentStart = line.length();
        if (stringStart == -1) stringStart = line.length();

        int min = Math.min(commentStart, Math.min(lineCommentStart, stringStart));

        if (min == lineCommentStart)
            return -1;

        if (min == stringStart) {
            int stringEnd = line.substring(stringStart + 1).indexOf("\"");
            if (stringEnd == -1) return -1;

            stringEnd += stringStart + 1;
            int comm = getCommentStart(line.substring(stringEnd + 1));

            return (comm == -1) ? (comm) : stringEnd + 1 + comm;
        }

        if (min == commentStart) {
            int commentEnd = line.substring(commentStart + 2).indexOf("*/");
            if (commentEnd == -1) return commentStart;

            commentEnd += commentStart + 2;
            int comm = getCommentStart(line.substring(commentEnd + 2));

            return (comm == -1) ? (comm) : commentEnd + 2 + comm;
        }

        return -1;
    }

    void updateComments(Element[] elems) {
        Element root = getDefaultRootElement();
        boolean commentStarted = false;

        int index = root.getElementIndex(elems[0].getStartOffset());
        if (index > 0) {
            Element elem = root.getElement(index - 1);
            AttributeSet attributes = elem.getAttributes();
            if (attributes.isDefined(CommentAttribute.comment)) {
                CommentAttribute comm = (CommentAttribute) attributes.getAttribute(CommentAttribute.comment);
                commentStarted = comm.isComment(elem.getEndOffset());
            }
        }

        for (Element elem : elems) {
            CommentAttribute comm = null;
            String line = null;

            try {
                line = getText(elem.getStartOffset(), elem.getEndOffset() - elem.getStartOffset());
            } catch (BadLocationException e) {
                // this should never happen
                logger.throwing("JavaDocument", "updateComments", e);
                continue;
            }

            if (commentStarted) {
                comm = new CommentAttribute();
                comm.setInComment(true);
            }

            boolean nextLineComment = commentStarted;

            if ((index = line.indexOf("*/")) != -1) {
                if (comm == null)
                    comm = new CommentAttribute();
                comm.setCommentEnd(index + 1);

                nextLineComment = false;

                int newCommentStart = getCommentStart(line.substring(index + 2));
                if (newCommentStart >= 0) {
                    comm.setNewCommentStart(index + 2 + newCommentStart);
                    if (commentStarted)
                        nextLineComment = true;
                }
            }

            int commentStart = getCommentStart(line);
            if (commentStart >= 0) {
                if (comm == null)
                    comm = new CommentAttribute();
                comm.setCommentStart(commentStart);
                if (!commentStarted)
                    nextLineComment = true;
            }

            commentStarted = nextLineComment;

            MutableAttributeSet attributes = (MutableAttributeSet) elem.getAttributes();
            attributes.removeAttribute(CommentAttribute.comment);
            if (comm != null)
                attributes.addAttribute(CommentAttribute.comment, comm);
        }

        index = root.getElementIndex(elems[elems.length - 1].getStartOffset());

        // see if the updated elements affects the next elements in the document
        while (true) {
            index ++;
            Element elem = root.getElement(index);
            if (elem == null) break;

            MutableAttributeSet attributes = (MutableAttributeSet) elem.getAttributes();
            if (attributes.isDefined(CommentAttribute.comment)) {
                CommentAttribute comm = (CommentAttribute) attributes.getAttribute(CommentAttribute.comment);
                if (commentStarted) {
                    comm.inComment = true;
                    if (comm.commentEnd >= 0) break;
                } else {
                    if (!comm.inComment) break;
                    comm.inComment = false;
                }
            } else {
                if (!commentStarted) break;
                CommentAttribute comm = new CommentAttribute();
                comm.setInComment(true);
                attributes.addAttribute(CommentAttribute.comment, comm);
            }
        }
    }

    public static class CommentAttribute {
        private int commentEnd;         // the first */ in the line
        private int commentStart;     // the first /* in the line that is not closed
                                     //        (assuming the line does not continue a comment)
        private int newCommentStart; // the first /* in the line that is not closed
                                     //     (assuming the line continues a comment)
        private boolean inComment;

        public static final String comment = "comment";

        public CommentAttribute() {
            commentEnd = commentStart = newCommentStart = -1;
            inComment = false;
        }

        public void setCommentEnd(int commentEnd) {
            this.commentEnd = commentEnd;
        }

        public void setCommentStart(int commentStart) {
            this.commentStart = commentStart;
        }

        public void setNewCommentStart(int newCommentStart) {
            this.newCommentStart = newCommentStart;
        }

        public void setInComment(boolean inComment) {
            this.inComment = inComment;
        }

        public boolean isComment(int pos) {
            return (inComment && (pos <= commentEnd || commentEnd < 0)) ||
                    (!inComment && (commentStart >= 0 && pos >= commentStart)) ||
                    (inComment && (newCommentStart >= 0 && pos >= newCommentStart));
        }

    }

    public static class Scanner extends sun.tools.java.Scanner {
        private boolean valid = false;
        private int start = 0;

        public Scanner() {
            super(new JavaDocument.Environment());
            scanComments = true;
        }

        public void restartScan(Document doc, int startPos) {
            if (!valid)
                try {
                    startPos = Math.min(startPos, doc.getLength() - 1);
                    Element root = doc.getDefaultRootElement();
                    startPos = root.getElement(root.getElementIndex(startPos)).getStartOffset();
                    useInputStream(new ByteArrayInputStream(doc.getText(startPos, doc.getLength() - startPos).getBytes()));
                    start = startPos;
                    valid = true;
                } catch (IOException e) {
                    logger.throwing("JavaDocument", "startScan", e);
                } catch (BadLocationException e) {
                    logger.throwing("JavaDocument", "startScan", e);
                }
            while (startPos >= (int) getEndPos()) {
                try {
                    scan();
                } catch (IOException e) {
                    logger.throwing("JavaDocument", "startScan", e);
                }
            }
        }

        public long getStartPos() {
            return start + pos;
        }

        public long getEndPos() {
            return start + super.getEndPos();
        }

        public void setValid(boolean b) {
            valid = b;
        }
    }

    public static class Environment extends sun.tools.java.Environment {
        public void error(Object source, long where, String err, Object arg1, Object arg2, Object arg3) {
        }
    }

    class JavaDocumentFilter extends DocumentFilter {

        private String filter(DocumentFilter.FilterBypass fb,
                int offset, String string) throws BadLocationException {

            PlainDocument doc = (PlainDocument) fb.getDocument();

            if (string.equals("\n") || string.equals("\r\n")) {

                int start = doc.getParagraphElement(offset).getStartOffset();
                String line = doc.getText(start, offset - start);

                StringBuffer buffer = new StringBuffer();
                for (int i = 0; i < line.length(); ++i)
                    if (line.charAt(i) == ' ')
                        buffer.append(' ');
                    else
                        break;

                string += buffer.toString();

            } else if (string.equals("\t")) {

                int start = doc.getParagraphElement(offset).getStartOffset();
                int cnt = (offset - start) % Config.TAB_SIZE;
                cnt = Config.TAB_SIZE - cnt;

                StringBuffer buffer = new StringBuffer();
                for (int i = 0; i < cnt; ++i)
                    buffer.append(' ');

                string = buffer.toString();

            }

            return string;

        }

        public void insertString(DocumentFilter.FilterBypass fb,
                int offset, String string, AttributeSet attr)
                throws BadLocationException {

            string = filter(fb, offset, string);
            super.insertString(fb, offset, string, attr);

        }

        public void replace(DocumentFilter.FilterBypass fb,
                int offset, int length, String text,
                AttributeSet attrs) throws BadLocationException {

            text = filter(fb, offset, text);
            super.replace(fb, offset, length, text, attrs);

        }

    }

}

