package language.java;

import java.io.*;
import java.io.File;
import java.util.*;
import java.util.List;
import java.util.jar.*;
import java.lang.reflect.*;
import com.sun.tools.javac.tree.*;
import com.sun.tools.javac.code.*;
import com.sun.tools.javac.util.*;
import com.sun.tools.javac.main.JavaCompiler;
import static com.sun.tools.javac.tree.JCTree.*;

import project.*;
import application.*;
import language.*;
import language.java.ccinfo.*;

import logging.Logging;
import static logging.Logging.logger;

public class JavaParser extends Thread {

    String filePath;
    int offset;
    String data = null;
    boolean localOnly;

    JCCompilationUnit tree = null;
    List <JCVariableDecl> visibleDecl = null;
    LinkedList <CCInfo> result = null;

    public JavaParser(String filePath, int offset, String data, boolean localOnly) {
        this.filePath = filePath;
        this.data = data;
        this.offset = offset;
        this.localOnly = localOnly;
    }

    public void run() {

        // Check if 'offset' is located in a comment area.
        int index = 0;
        while (true) {
            index = data.indexOf("//", index);
            if (index == -1 || index >= offset)
                break;
            index = data.indexOf('\n', index);
            if (index == -1 || offset < index)
                return;
        }
        index = 0;
        while (true) {
            index = data.indexOf("/*", index);
            if (index == -1 || index >= offset)
                break;
            index = data.indexOf("*/", index + 2);
            if (index == -1 || offset < index)
                return;
        }

        // Compile the document and get an AST.
        compileCurrentDocument();
        if (tree == null)
            return;

        result = new LinkedList <CCInfo> ();

        // Get only the local visible variables.
        if (localOnly) {
            visibleDecl = new LinkedList <JCVariableDecl> ();
            listVisibleDecl(tree);
            for (JCVariableDecl decl : visibleDecl)
                result.addFirst(new FieldInfo(decl, null));
            CCPopup.setList(result);
            return;
        }

        String name = getDotSeparatedName(offset);
        int pos = getDotSeparatedStartPos(offset);
        if (name == null || name.equals(""))
            return;

        int ppos = readBackward("import", pos - 1);
        if (ppos != -1) {
            // This request is for an 'import'.
            getImportSgg(name);
            CCPopup.setList(result);
            return;
        } else {
            ppos = readBackward("static", pos - 1);
            if (ppos != -1)
                ppos = readBackward("import", ppos - 1);
            if (ppos != -1) {
                // This request is for an 'import static'.
                getStaticMembersSgg(name);
                getImportSgg(name);
                CCPopup.setList(result);
                return;
            }
        }

        visibleDecl = new LinkedList <JCVariableDecl> ();
        listVisibleDecl(tree);
        getNameSgg(getDotSeparatedName(offset), offset);
        CCPopup.setList(result);

    }

    private void compileCurrentDocument () {
        try {
            // Create a temporary file.
            File tmp = File.createTempFile(
                    "java-parser-" + System.currentTimeMillis(),
                    ".java");

            // Write data to that file.
            FileOutputStream out = new FileOutputStream(tmp);
            out.write(data.getBytes());
            out.close();

            // Redirect System.out and System.err to the logger.
            Logging.startLoggingSystemOut("language.java.JavaParser", "run");
            Logging.startLoggingSystemErr("language.java.JavaParser", "run");

            // Parse the file for getting a tree.
            Context context = new Context();
            Options options = Options.instance(context);
            options.put("-Xjcov", ""); // option for end positions
            JavaCompiler compiler = new JavaCompiler(context);
            tree = compiler.parse(tmp.getAbsolutePath()); // FIXME: deprecated

            // Restore System.out and System.err.
            Logging.stopLoggingSystemOut();
            Logging.stopLoggingSystemErr();

            // Delete the temporary file.
            tmp.delete();

        } catch (IOException e) {
            logger.throwing("language.java.JavaParser", "run", e);
            return;
        }
    }

    private boolean checkInside(int offset, JCTree t) {
        return t != null &&
            t.getStartPosition() <= offset &&
            offset < t.getEndPosition(tree.endPositions);
    }

    private boolean checkAfter(int offset, JCTree t) {
        return t != null &&
            offset >= t.getEndPosition(tree.endPositions);
    }

    private void listVisibleDecl(JCTree t) {

        if (t == null)
            return;

        boolean ok;
        JCStatement body;

        switch (t.tag) {

            case TOPLEVEL:
                for (JCTree def : ((JCCompilationUnit) t).defs)
                    if (def.tag == CLASSDEF && checkInside(offset, def))
                        listVisibleDecl(def);
                break;

            case CLASSDEF:
                for (JCTree def : ((JCClassDecl) t).defs)
                    if (def.tag == VARDEF)
                        listVisibleDecl(def);
                for (JCTree def : ((JCClassDecl) t).defs)
                    if ((def.tag == METHODDEF || def.tag == CLASSDEF)
                            && checkInside(offset, def))
                        listVisibleDecl(def);
                break;

            case METHODDEF:
                for (JCTree def : ((JCMethodDecl) t).params)
                    listVisibleDecl(def);
                if (checkInside(offset, ((JCMethodDecl) t).body))
                    listVisibleDecl(((JCMethodDecl) t).body);
                break;

            case VARDEF:
                if (!checkInside(offset, t))
                    visibleDecl.add((JCVariableDecl) t);
                break;

            case BLOCK:
                for (JCStatement st : ((JCBlock) t).stats)
                    if (checkInside(offset, st)) {
                        listVisibleDecl(st);
                        break;
                    } else if (checkAfter(offset, st) && st instanceof JCVariableDecl) {
                        listVisibleDecl(st);
                    }
                break;

            case DOLOOP:
                body = ((JCDoWhileLoop) t).body;
                if (checkInside(offset, body))
                    listVisibleDecl(body);
                break;

            case WHILELOOP:
                body = ((JCWhileLoop) t).body;
                if (checkInside(offset, body)) {
                    listVisibleDecl(body);
                }
                break;

            case FORLOOP:
                ok = true;
                for (JCStatement st : ((JCForLoop) t).init)
                    if (checkInside(offset, st)) {
                        ok = false;
                        break;
                    } else if (st instanceof JCVariableDecl)
                        listVisibleDecl(st);
                if (!ok)
                    break;

                body = ((JCForLoop) t).body;
                if (checkInside(offset, body))
                    listVisibleDecl(body);
                break;

            case FOREACHLOOP:
                listVisibleDecl(((JCEnhancedForLoop) t).var);
                body = ((JCEnhancedForLoop) t).body;
                if (checkInside(offset, body))
                    listVisibleDecl(body);
                break;

            case LABELLED:
                body = ((JCLabeledStatement) t).body;
                if (checkInside(offset, body))
                    listVisibleDecl(body);
                break;

            case SWITCH:
                ok = false;
                for (JCCase c : ((JCSwitch) t).cases)
                    if (checkInside(offset, c))
                        ok = true;
                if (!ok)
                    break;

                for (JCCase c : ((JCSwitch) t).cases) {
                    listVisibleDecl(c);
                    if (checkInside(offset, c))
                        break;
                }
                break;

            case CASE:
                for (JCStatement st : ((JCCase) t).stats)
                    if (checkInside(offset, st)) {
                        listVisibleDecl(st);
                        break;
                    } else if (st instanceof JCVariableDecl)
                        listVisibleDecl(st);
                break;

            case SYNCHRONIZED:
                body = ((JCSynchronized) t).body;
                if (checkInside(offset, body))
                    listVisibleDecl(body);
                break;

            case TRY:
                if (checkInside(offset, ((JCTry) t).body))
                    listVisibleDecl(((JCTry) t).body);
                for (JCCatch c : ((JCTry) t).catchers)
                    if (checkInside(offset, c))
                        listVisibleDecl(c);
                if (checkInside(offset, ((JCTry) t).finalizer))
                    listVisibleDecl(((JCTry) t).finalizer);
                break;

            case CATCH:
                listVisibleDecl(((JCCatch) t).param);
                if (checkInside(offset, ((JCCatch) t).body))
                    listVisibleDecl(((JCCatch) t).body);
                break;

            case IF:
                if (checkInside(offset, ((JCIf) t).thenpart))
                    listVisibleDecl(((JCIf) t).thenpart);
                if (checkInside(offset, ((JCIf) t).elsepart))
                    listVisibleDecl(((JCIf) t).elsepart);
                break;

        }

    }

    private String getDotSeparatedName(int pos) {
        StringBuffer out = new StringBuffer();

        --pos;

        while (true) {
            char ch;

            // spaces after name
            while (pos >= 0) {
                ch = data.charAt(pos);
                if (!Character.isSpaceChar(ch))
                    break;
                --pos;
            }

            // name
            boolean found = false;
            while (pos >= 0) {
                ch = data.charAt(pos);
                if (!Character.isJavaIdentifierPart(ch))
                    break;
                found = true;
                out.append(ch);
                --pos;
            }
            if (!found)
                return null;

            // spaces before name
            while (pos >= 0) {
                ch = data.charAt(pos);
                if (!Character.isSpaceChar(ch))
                    break;
                --pos;
            }

            // dot
            ch = data.charAt(pos);
            if (ch != '.')
                break;

            out.append('.');
            --pos;
        }

        String result = out.reverse().toString();
        if (result.charAt(0) == '.')
            return null;
        return result;
    }

    private int getDotSeparatedStartPos(int pos) {
        int firstPos = -1;

        --pos;

        while (true) {
            char ch;

            // spaces after name
            while (pos >= 0) {
                ch = data.charAt(pos);
                if (!Character.isSpaceChar(ch))
                    break;
                --pos;
            }

            // name
            boolean found = false;
            while (pos >= 0) {
                ch = data.charAt(pos);
                if (!Character.isJavaIdentifierPart(ch))
                    break;
                else
                    firstPos = pos;
                found = true;
                --pos;
            }
            if (!found)
                return -1;

            // spaces before name
            while (pos >= 0) {
                ch = data.charAt(pos);
                if (!Character.isSpaceChar(ch))
                    break;
                --pos;
            }

            // dot
            ch = data.charAt(pos);
            if (ch != '.')
                break;

            --pos;
        }

        return firstPos;
    }

    private int readBackward(String str, int pos) {
        while (pos >= 0) {
            char ch = data.charAt(pos);
            if (!Character.isSpaceChar(ch))
                break;
            --pos;
        }
        if (pos - str.length() + 1 < 0)
            return -1;
        if (data.substring(pos - str.length() + 1, pos + 1).equals(str))
            return pos - str.length() + 1;
        else
            return -1;
    }

    // Given the full name of a package, get the list of subpackages and classes.
    private void getImportSgg(String name) {

        // Search in the current project structure.
        String dir = ProjectManager.getProjectPathForFile(filePath) + name.replace('.', '/');
        File f = new File(dir);
        if (f.exists() && f.isDirectory()) {
            for (String str : f.list()) {
                File ff = new File(dir + "/" + str);
                if (ff.isDirectory())
                    result.add(new ImportInfo(str));
                else if (str.endsWith(".java"))
                    result.add(new ImportInfo(str.substring(0, str.length() - 5)));
            }
        }

        // Search in the Java language.
        Set <String> set = new TreeSet <String> ();

        dir = System.getProperty("java.home") + "/lib";
        f = new File(dir);
        for (String str : f.list())
            if (str.endsWith(".jar"))
                searchInJar(name, dir + "/" + str, set);

        for (String dir2 : System.getProperty("java.ext.dirs").split(System.getProperty("path.separator"))) {
            f = new File(dir2);
            if (!f.exists() || !f.isDirectory())
                continue;
            for (String str : f.list())
                if (str.endsWith(".jar"))
                    searchInJar(name, dir2 + "/" + str, set);
        }

        for (String dir2 : System.getProperty("java.class.path").split(System.getProperty("path.separator"))) {
            f = new File(dir2);
            if (!f.exists() || !f.isDirectory())
                continue;
            for (String str : f.list())
                if (str.endsWith(".jar"))
                    searchInJar(name, dir2 + "/" + str, set);
        }

        for (String entry : set)
            result.add(new ImportInfo(entry));

    }

    // Given the full name of a class, get the list of static fields and classes.
    private void getStaticMembersSgg(String name) {

        // Search in Java libraries.
        try {
            Class <?> c = Class.forName(name);

            for (Class <?> cls : c.getClasses()) {
                int mod = cls.getModifiers();
                if (Modifier.isStatic(mod) && Modifier.isPublic(mod))
                    result.add(new ClassInfo(cls));
            }

            for (Field f : c.getFields()) {
                int mod = f.getModifiers();
                if (Modifier.isStatic(mod) && Modifier.isPublic(mod))
                    result.add(new FieldInfo(f));
            }

        } catch (ClassNotFoundException e) {
            logger.finer("Class not found (CC): " + e.getMessage());
        }

        // Search in current project.
        String projectPath = ProjectManager.getProjectPathForFile(filePath);
        String className = name.substring(name.lastIndexOf('.') + 1);
        String jpath = projectPath + name.replace('.', '/') + ".java";


        File f = new File(jpath);
        if (f.exists() && !f.isDirectory()) {

            // Redirect System.out and System.err to the logger.
            Logging.startLoggingSystemOut("language.java.JavaParser", "getStaticMembersSgg");
            Logging.startLoggingSystemErr("language.java.JavaParser", "getStaticMembersSgg");

            // Parse the file for getting a tree.
            JavaCompiler compiler = new JavaCompiler(new Context());
            try {
                JCTree.JCCompilationUnit treeFile = compiler.parse(jpath); // FIXME: deprecated

                for (JCTree t : treeFile.defs)
                    if (t.tag == CLASSDEF && className.equals(((JCClassDecl) t).getSimpleName().toString())) {

                        for (JCTree tt : ((JCClassDecl) t).defs)
                            if (tt.tag == CLASSDEF) {
                                JCClassDecl cls = (JCClassDecl) tt;
                                long mod = cls.mods.flags;
                                if ((mod & Flags.STATIC) != 0 && (mod & Flags.PUBLIC) != 0)
                                    result.add(new ClassInfo(cls, name));
                            }

                        for (JCTree tt : ((JCClassDecl) t).defs)
                            if (tt.tag == VARDEF) {
                                JCVariableDecl var = (JCVariableDecl) tt;
                                long mod = var.mods.flags;
                                if ((mod & Flags.STATIC) != 0 && (mod & Flags.PUBLIC) != 0)
                                    result.add(new FieldInfo(var, name));
                            }

                        break;
                    }

            } catch (IOException e) {
                logger.throwing("language.java.JavaParser", "getStaticMembersSgg", e);
            }

            // Restore System.out and System.err.
            Logging.stopLoggingSystemOut();
            Logging.stopLoggingSystemErr();

        }

    }

    private void searchInJar(String name, String jar, Set <String> set) {
        try {
            JarFile file = new JarFile(jar);
            Enumeration <JarEntry> entries = file.entries();
            int cnt = name.split("\\.").length;
            String str;

            name = name.replace('.', '/') + "/";

            while (entries.hasMoreElements()) {

                str = entries.nextElement().getName();

                if (str.startsWith(name) && str.indexOf('$') == -1) {

                    String[] parts = str.split("/");

                    if (parts.length >= cnt + 1) {
                        if (parts[cnt].indexOf('.') == -1)
                            set.add(parts[cnt]);
                        else if (parts[cnt].endsWith(".class"))
                            set.add(parts[cnt].substring(0, parts[cnt].length() - 6));
                    }

                }

            }

        } catch (IOException e) {
            logger.throwing("language.java.JavaParser", "searchInJar", e);
            return;
        }
    }

    // Given a dot separated name and the position of the request, get specific suggestions.
    private void getNameSgg(String name, int pos) {

        if (name.equals("this")) {
            getThisSgg(tree, pos);
            return;
        }

        JCVariableDecl declaration = null;
        for (JCVariableDecl var : visibleDecl)
            if (var.getName().toString().equals(name))
                declaration = var;

        // List static things.
        if (declaration == null) {
            if (getInnerClassSgg(tree, pos, name, true))
                return;
            if (getClassSgg(tree, name, true))
                return;
            if (getLocalImportSgg(tree, name, true))
                return;
            if (getLocalPackageSgg(name, true))
                return;
            if (getStarImportSgg(tree, name, true))
                return;
            if (getJavaLangSgg(name, true))
                return;
            getStaticMembersSgg(name);
            getImportSgg(name);
            return;
        }

        name = declaration.getType().toString();
        if (getInnerClassSgg(tree, pos, name, false))
            return;
        if (getClassSgg(tree, name, false))
            return;
        if (getLocalImportSgg(tree, name, false))
            return;
        if (getLocalPackageSgg(name, false))
            return;
        if (getStarImportSgg(tree, name, false))
            return;
        if (getJavaLangSgg(name, false))
            return;
        getFullNameClassSgg(name, false);

    }

    // Check first if the name is for an inner class, i.e. we must prefix it
    // with the name of the enclosing class.
    private boolean getInnerClassSgg(JCCompilationUnit tree, int pos,
            String name, boolean isStatic) {

        List <String> classes = new LinkedList <String> ();
        List <JCTree> defs = tree.defs;

        while (true) {
            boolean found = false;
            for (JCTree t : defs)
                if (t.tag == CLASSDEF && checkInside(pos, t)) {
                    found = true;
                    classes.add(((JCClassDecl) t).getSimpleName().toString());
                    defs = ((JCClassDecl) t).defs;
                }
            if (!found)
                break;
        }


        String prefix = "";
        for (String str : classes) {
            prefix += str + ".";
            if (getClassSgg(tree, prefix + name, isStatic))
                return true;
        }

        return false;

    }

    // Search for a class in the parse tree. The name of the class is a dot
    // separated name.
    private boolean getClassSgg(JCCompilationUnit tree, String name,
            boolean isStatic) {

        List <JCTree> defs = tree.defs;
        JCClassDecl declaringClass = null;
        String[] parts = name.split("\\.");

        for (int i = 0; i < parts.length; ++i) {
            boolean found = false;
            for (JCTree t : defs)
                if (t.tag == CLASSDEF && parts[i].equals(((JCClassDecl) t).getSimpleName().toString())) {
                    found = true;
                    declaringClass = (JCClassDecl) t;
                    defs = declaringClass.defs;
                    break;
                }
            if (found == false)
                return false;
        }

        // List classes.
        for (JCTree t : defs)
            if (t.tag == CLASSDEF) {
                JCClassDecl cls = (JCClassDecl) t;
                long clsFlags = cls.mods.flags;
                if ((clsFlags & Flags.PUBLIC) != 0 &&
                        (!isStatic || (clsFlags & Flags.STATIC) != 0))
                    result.add(new ClassInfo(cls,
                            (declaringClass == null ? null : declaringClass.getSimpleName().toString())));
            }

        // List fields.
        for (JCTree t : defs)
            if (t.tag == VARDEF) {
                JCVariableDecl var = (JCVariableDecl) t;
                long varFlags = var.mods.flags;
                if ((varFlags & Flags.PUBLIC) != 0 &&
                        (!isStatic || (varFlags & Flags.STATIC) != 0))
                    result.add(new FieldInfo(var,
                            (declaringClass == null ? null : declaringClass.getSimpleName().toString())));
            }

        // List methods.
        for (JCTree t : defs)
            if (t.tag == METHODDEF) {
                JCMethodDecl mth = (JCMethodDecl) t;
                long mthFlags = mth.mods.flags;
                if ((mthFlags & Flags.PUBLIC) != 0 &&
                        (!isStatic || (mthFlags & Flags.STATIC) != 0))
                    result.add(new MethodInfo(mth,
                            (declaringClass == null ? null : declaringClass.getSimpleName().toString())));
            }

        return true;

    }

    // Combine the name with the non-star imports and get suggestions.
    private boolean getLocalImportSgg(JCCompilationUnit tree, String name,
            boolean isStatic) {

        String[] tmp = name.split("\\.", 2);
        String partA = tmp[0];
        String partB = (tmp.length < 2) ? null : tmp[1];

        for (JCTree t : tree.defs)
            if (t.tag == IMPORT) {
                JCImport imp = (JCImport) t;
                if (imp.isStatic())
                    continue;
                String impName = imp.getQualifiedIdentifier().toString();
                if (impName.endsWith("." + partA)) {
                    getFullNameClassSgg(impName + (partB == null ? "" : "." + partB), isStatic);
                    return true;
                }
            }

        return false;

    }

    // Get suggestions for the classes in the same package as the current file.
    private boolean getLocalPackageSgg(String name, boolean isStatic) {

        int index = filePath.lastIndexOf('/');
        if (index == -1)
            index = filePath.length();
        String currentFolder = filePath.substring(0, index);

        if (name.indexOf('.') != -1)
            return false;

        String jpath = currentFolder + "/" + name + ".java";

        File f = new File(jpath);
        if (f.exists() && !f.isDirectory()) {

            // Redirect System.out and System.err to the logger.
            Logging.startLoggingSystemOut("language.java.JavaParser", "getLocalPackageSgg");
            Logging.startLoggingSystemErr("language.java.JavaParser", "getLocalPackageSgg");

            // Parse the file for getting a tree.
            JavaCompiler compiler = new JavaCompiler(new Context());
            try {
                JCTree.JCCompilationUnit treeFile = compiler.parse(jpath); // FIXME: deprecated

                for (JCTree t : treeFile.defs)
                    if (t.tag == CLASSDEF && name.equals(((JCClassDecl) t).getSimpleName().toString())) {

                        for (JCTree tt : ((JCClassDecl) t).defs)
                            if (tt.tag == CLASSDEF) {
                                JCClassDecl cls = (JCClassDecl) tt;
                                long mod = cls.mods.flags;
                                if ((mod & Flags.PUBLIC) != 0 && (!isStatic || (mod & Flags.STATIC) != 0))
                                    result.add(new ClassInfo(cls, name));
                            }

                        for (JCTree tt : ((JCClassDecl) t).defs)
                            if (tt.tag == VARDEF) {
                                JCVariableDecl var = (JCVariableDecl) tt;
                                long mod = var.mods.flags;
                                if ((mod & Flags.PUBLIC) != 0 && (!isStatic || (mod & Flags.STATIC) != 0))
                                    result.add(new FieldInfo(var, name));
                            }

                        for (JCTree tt : ((JCClassDecl) t).defs)
                            if (tt.tag == METHODDEF) {
                                JCMethodDecl mth = (JCMethodDecl) tt;
                                long mod = mth.mods.flags;
                                if ((mod & Flags.PUBLIC) != 0 && (!isStatic || (mod & Flags.STATIC) != 0))
                                    result.add(new MethodInfo(mth, name));
                            }

                        return true;
                    }

            } catch (IOException e) {
                logger.throwing("language.java.JavaParser", "getLocalPackageSgg", e);
            }

            // Restore System.out and System.err.
            Logging.stopLoggingSystemOut();
            Logging.stopLoggingSystemErr();

        }

        return false;

    }

    // Combine the name with star imports and get suggestions.
    private boolean getStarImportSgg(JCCompilationUnit tree, String name,
            boolean isStatic) {

        if (name.indexOf('.') != -1)
            return false;

        for (JCTree t : tree.defs)
            if (t.tag == IMPORT) {
                JCImport imp = (JCImport) t;
                String impName = imp.getQualifiedIdentifier().toString();
                if (impName.endsWith("*"))
                    if (getFullNameClassSgg(impName.substring(0, impName.length() - 1)
                            + name, isStatic))
                        return true;
            }

        return false;

    }

    // Get suggestions for the packet java.lang
    private boolean getJavaLangSgg(String name, boolean isStatic) {
        return getFullNameClassSgg("java.lang." + name, isStatic);
    }

    // Pretty much the same that getStaticMembersSgg function does.
    private boolean getFullNameClassSgg(String name, boolean isStatic) {

        // Search in Java libraries.
        try {
            Class <?> c = Class.forName(name);

            for (Class <?> cls : c.getClasses()) {
                int mod = cls.getModifiers();
                if (Modifier.isPublic(mod) && (!isStatic || Modifier.isStatic(mod)))
                    result.add(new ClassInfo(cls));
            }

            for (Field f : c.getFields()) {
                int mod = f.getModifiers();
                if (Modifier.isPublic(mod) && (!isStatic || Modifier.isStatic(mod)))
                    result.add(new FieldInfo(f));
            }

            for (Method m : c.getMethods()) {
                int mod = m.getModifiers();
                if (Modifier.isPublic(mod) && (!isStatic || Modifier.isStatic(mod)))
                    result.add(new MethodInfo(m));
            }

            return true;

        } catch (ClassNotFoundException e) {
            logger.finer("Class not found (CC): " + e.getMessage());
        }

        // Search in current project.
        String projectPath = ProjectManager.getProjectPathForFile(filePath);
        String className = name.substring(name.lastIndexOf('.') + 1);
        String jpath = projectPath + name.replace('.', '/') + ".java";

        File f = new File(jpath);
        if (f.exists() && !f.isDirectory()) {

            // Redirect System.out and System.err to the logger.
            Logging.startLoggingSystemOut("language.java.JavaParser", "getFullNameClassSgg");
            Logging.startLoggingSystemErr("language.java.JavaParser", "getFullNameClassSgg");

            // Parse the file for getting a tree.
            JavaCompiler compiler = new JavaCompiler(new Context());
            try {
                JCTree.JCCompilationUnit treeFile = compiler.parse(jpath); // FIXME: deprecated

                for (JCTree t : treeFile.defs)
                    if (t.tag == CLASSDEF && className.equals(((JCClassDecl) t).getSimpleName().toString())) {

                        for (JCTree tt : ((JCClassDecl) t).defs)
                            if (tt.tag == CLASSDEF) {
                                JCClassDecl cls = (JCClassDecl) tt;
                                long mod = cls.mods.flags;
                                if ((mod & Flags.PUBLIC) != 0 && (!isStatic || (mod & Flags.STATIC) != 0))
                                    result.add(new ClassInfo(cls, name));
                            }

                        for (JCTree tt : ((JCClassDecl) t).defs)
                            if (tt.tag == VARDEF) {
                                JCVariableDecl var = (JCVariableDecl) tt;
                                long mod = var.mods.flags;
                                if ((mod & Flags.PUBLIC) != 0 && (!isStatic || (mod & Flags.STATIC) != 0))
                                    result.add(new FieldInfo(var, name));
                            }

                        for (JCTree tt : ((JCClassDecl) t).defs)
                            if (tt.tag == METHODDEF) {
                                JCMethodDecl mth = (JCMethodDecl) tt;
                                long mod = mth.mods.flags;
                                if ((mod & Flags.PUBLIC) != 0 && (!isStatic || (mod & Flags.STATIC) != 0))
                                    result.add(new MethodInfo(mth, name));
                            }

                        return true;
                    }

            } catch (IOException e) {
                logger.throwing("language.java.JavaParser", "getFullNameClassSgg", e);
            }

            // Restore System.out and System.err.
            Logging.stopLoggingSystemOut();
            Logging.stopLoggingSystemErr();

        }

        return false;

    }

    private void getThisSgg(JCCompilationUnit tree, int pos) {
        JCClassDecl c = searchLastClass(tree, pos);
        if (c != null) {
            for (JCTree def : c.defs)
                if (def.tag == CLASSDEF)
                    result.add(new ClassInfo((JCClassDecl) def, c.getSimpleName().toString()));
            for (JCTree def : c.defs)
                if (def.tag == VARDEF)
                    result.add(new FieldInfo((JCVariableDecl) def, c.getSimpleName().toString()));
            for (JCTree def : c.defs)
                if (def.tag == METHODDEF)
                    result.add(new MethodInfo((JCMethodDecl) def, c.getSimpleName().toString()));
        }
    }

    private JCClassDecl searchLastClass(JCTree tree, int pos) {
        JCClassDecl c = null;

        switch (tree.tag) {
        case TOPLEVEL:
            for (JCTree def : ((JCCompilationUnit) tree).defs)
                if (def.tag == CLASSDEF && checkInside(pos, def))
                    c = searchLastClass(def, pos);
            break;
        case CLASSDEF:
            for (JCTree def : ((JCClassDecl) tree).defs)
                if (def.tag == CLASSDEF && checkInside(pos, def))
                    c = searchLastClass(def, pos);
            if (c == null)
                c = (JCClassDecl) tree;
            break;
        }

        return c;
    }

}
