package language.xml;

import java.awt.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentEvent.ElementChange;
import javax.swing.text.*;

import language.xml.XMLDocument.CommentAttribute;

public class XMLContext extends StyleContext implements ViewFactory {

    public XMLContext() {
        super();
    }

    public View create(Element elem) {
        return new XMLView(elem);
    }

    static class XMLView extends WrappedPlainView {

        private static TreeMap<Integer, Pattern> patternTokens;
        private static final String TAG_CDATA_START = "(\\<!\\[CDATA\\[).*";
        private static final String TAG_CDATA_END = ".*(]]>)";
        private static final String TAG_PATTERN = "(</?\\w*)\\s?>?";
        private static final String TAG_ATTRIBUTE_PATTERN = "(^|\\s)(\\w*)\\=";
        private static final String TAG_END_PATTERN = "(>|/>)";
        private static final String TAG_ATTRIBUTE_VALUE = "\\w*\\=(\"[^\"]*\")";
        private static final String TAG_COMMENT = "(<!--.*(-->)?)";

        {
            patternTokens = new TreeMap<Integer, Pattern>();
            patternTokens.put(XMLToken.TAG_CDATA_START, Pattern.compile(TAG_CDATA_START));
            patternTokens.put(XMLToken.TAG_CDATA_END, Pattern.compile(TAG_CDATA_END));
            patternTokens.put(XMLToken.TAG_PATTERN, Pattern.compile(TAG_PATTERN));
            patternTokens.put(XMLToken.TAG_ATTRIBUTE_PATTERN, Pattern.compile(TAG_ATTRIBUTE_PATTERN));
            patternTokens.put(XMLToken.TAG_END_PATTERN, Pattern.compile(TAG_END_PATTERN));
            patternTokens.put(XMLToken.TAG_ATTRIBUTE_VALUE, Pattern.compile(TAG_ATTRIBUTE_VALUE));
            patternTokens.put(XMLToken.TAG_COMMENT, Pattern.compile(TAG_COMMENT));
        }

        public XMLView(Element element) {
            super(element, true);
        }

        public void insertUpdate(DocumentEvent chng, Shape a, ViewFactory f) {
            super.insertUpdate(chng, a, f);

            // updates elements comment attributes
            XMLDocument doc = (XMLDocument) getDocument();
            Element root = doc.getDefaultRootElement();
            ElementChange changes = chng.getChange(root);

            if (changes != null) {
                Element[] elems = changes.getChildrenAdded();
                doc.updateComments(elems);
            } else {
                Element[] elems = new Element[1];
                elems[0] = root.getElement(root.getElementIndex(chng.getOffset()));
                doc.updateComments(elems);
            }

            getContainer().repaint();
        }

        public void removeUpdate(DocumentEvent chng, Shape a, ViewFactory f) {
                                                                        super.removeUpdate(chng, a, f);

            // updates elements comment attributes
            XMLDocument doc = (XMLDocument) getDocument();
            Element root = doc.getDefaultRootElement();
            Element[] elems = new Element[1];
            elems[0] = root.getElement(root.getElementIndex(chng.getOffset()));
            doc.updateComments(elems);

            getContainer().repaint();
        }

        protected int drawUnselectedText(Graphics g, int x, int y, int startPos, int endPos)
                                                                                throws BadLocationException {
            return draw(g, x, y, startPos, endPos, true);
        }

        protected int drawSelectedText(Graphics g, int x, int y, int startPos, int endPos)
                                                                                throws BadLocationException {
            return draw(g, x, y, startPos, endPos, false);
        }

        private int draw(Graphics g, int x, int y, int startPos, int endPos, boolean withColors)
                                                                                    throws BadLocationException {
            Document doc = getDocument();
            Element root = doc.getDefaultRootElement();

            int startScan = root.getElement(root.getElementIndex(startPos)).getStartOffset();
            int endScan = root.getElement(root.getElementIndex(endPos)).getEndOffset();
            String text = doc.getText(startScan, endScan - startScan);

            TreeMap<Integer, XMLToken> tokenMap = new TreeMap<Integer, XMLToken>();
            for (Map.Entry<Integer, Pattern> entry : patternTokens.entrySet()) {
                Matcher matcher = entry.getValue().matcher(text);
                while (matcher.find())
                    tokenMap.put(matcher.start(1), new XMLToken(matcher.end(), entry.getKey()));
            }

            Segment segment = getLineBuffer();

            Font font = g.getFont();
            font = new Font(font.getFamily(), Font.PLAIN, font.getSize());
            if (!withColors)
                g.setColor(Color.WHITE);

            XMLToken lastToken = null;

            for (Map.Entry<Integer, XMLToken> entry : tokenMap.entrySet()) {
                if (startPos >= endPos) break;

                XMLToken token = entry.getValue();
                int start = startScan + entry.getKey();
                int end = startScan + token.getEndPos();

                // see if we are in a comment
                int endComment = startPos;
                while (endComment < endPos) {
                    Element elem = root.getElement(root.getElementIndex(endComment));
                    if (elem != null) {
                        AttributeSet attributes = elem.getAttributes();
                        if (attributes.isDefined(language.xml.XMLDocument.CommentAttribute.comment) &&
                                ((CommentAttribute) attributes.getAttribute(CommentAttribute.comment)).
                                                    isComment(endComment - elem.getStartOffset())) {
                            endComment++;
                            continue;
                        }
                    }
                    break;
                }

                // draw comment
                if (startPos < endComment) {
                    XMLToken comment = new XMLToken(startPos, XMLToken.TAG_COMMENT);
                    if (withColors) g.setColor(comment.getColor());
                    g.setFont(new Font(font.getFamily(), comment.getFontStyle(), font.getSize()));
                    doc.getText(startPos, endComment - startPos, segment);
                    x = Utilities.drawTabbedText(segment, x, y, g, this, startPos);
                    startPos = endComment;
                }

                if (end <= startPos || start >= endPos || (lastToken != null && start < lastToken.getEndPos()))
                    continue;

                // draw text element
                if (startPos < start) {
                    if (withColors) g.setColor(Color.BLACK);
                    g.setFont(font);
                    doc.getText(startPos, start - startPos, segment);
                    x = Utilities.drawTabbedText(segment, x, y, g, this, startPos);
                    startPos = start;
                }

                // draw curent token
                if (withColors) g.setColor(token.getColor());
                g.setFont(new Font(font.getFamily(), token.getFontStyle(), font.getSize()));
                end = Math.min(end, endPos);
                doc.getText(startPos, end - startPos, segment);
                x = Utilities.drawTabbedText(segment, x, y, g, this, startPos);
                startPos = end;
            }

            if (startPos < endPos) {
                // update comments if necessary
                int endComment = startPos;
                while (endComment < endPos) {
                    Element elem = root.getElement(root.getElementIndex(endComment));
                    if (elem != null) {
                        AttributeSet attributes = elem.getAttributes();
                        if (attributes.isDefined(language.xml.XMLDocument.CommentAttribute.comment) &&
                                ((CommentAttribute) attributes.getAttribute(CommentAttribute.comment)).
                                                    isComment(endComment - elem.getStartOffset())) {
                            endComment++;
                            continue;
                        }
                    }
                    break;
                }

                // draw comment
                if (startPos < endComment) {
                    XMLToken comment = new XMLToken(startPos, XMLToken.TAG_COMMENT);
                    if (withColors) g.setColor(comment.getColor());
                    g.setFont(new Font(font.getFamily(), comment.getFontStyle(), font.getSize()));
                    doc.getText(startPos, endComment - startPos, segment);
                    x = Utilities.drawTabbedText(segment, x, y, g, this, startPos);
                    startPos = endComment;
                }

                // draw remaining text
                if (startPos < endPos) {
                    if (withColors) g.setColor(Color.BLACK);
                    g.setFont(font);
                    doc.getText(startPos, endPos - startPos, segment);
                    x = Utilities.drawTabbedText(segment, x, y, g, this, startPos);
                }
            }

            return x;
        }

    }
}
