package plugins;

import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.text.AttributeSet;
import javax.swing.text.Document;
import javax.swing.text.AbstractDocument.DefaultDocumentEvent;

import java.awt.event.*;
import application.*;
import java.util.List;
import static logging.Logging.logger;

public class PluginManager {

    private static List <Plugin> plugins = new LinkedList <Plugin> ();

    public static void init() {
        try {
            File dir = new File("plugins");
            if (!dir.exists()) {
                dir.mkdir();
                dir.createNewFile();
            }
            for (File file : dir.listFiles())
                if (file.getName().endsWith(".class")) {

                    // Get the name of the class.
                    String name = file.getName().substring(0,
                            file.getName().indexOf(".class"));

                    // Load the class and test if it is a plugin.
                    try {
                        Class p = ClassLoader.getSystemClassLoader().loadClass("plugins." + name);
                        try {
                            Object o = p.newInstance();
                            if (o instanceof Plugin) {
                                Plugin plugin = (Plugin) o;
                                plugins.add(plugin);
                                if (isActivePlugin(name)) {
                                    plugin.init();
                                }
                            }
                        } catch (InstantiationException e) {
                            logger.info("Instantiation exception: " + e.getMessage());
                        } catch (IllegalAccessException e) {
                            logger.info("Illegal access: " + e.getMessage());
                        }

                    } catch (ClassNotFoundException e) {
                        logger.info("Class not found: " + e.getMessage());
                    }

                    Application.updatePluginsMenu();

                }
        } catch (IOException e) {
            logger.throwing("plugins.PluginManager", "init", e);
        }
    }

    public static boolean isActivePlugin(String name) {
        return Config.isActivePlugin(name);
    }

    public static void setActivatedPlugin(String name, boolean isActivated) {
        for (Plugin p : plugins)
            if (name.equals(p.getName())) {
                if (isActivated) {
                    if (!isActivePlugin(name))
                        p.init();
                } else {
                    if (isActivePlugin(name))
                        p.clear();
                }
                break;
            }
        Config.setActivatedPlugin(name, isActivated);
    }

    public static void insertUpdate(Document doc, DefaultDocumentEvent chng,
            AttributeSet attr) {
        for (Plugin p : plugins)
            if (isActivePlugin(p.getName()))
                p.insertUpdate(doc, chng, attr);
    }

    public static void removeUpdate(Document doc, DefaultDocumentEvent chng) {
        for (Plugin p : plugins)
            if (isActivePlugin(p.getName()))
                p.removeUpdate(doc, chng);
    }

    public static JMenu getJMenu() {
        JMenu menu = new JMenu("Plugins");
        menu.setMnemonic(java.awt.event.KeyEvent.VK_P);
        JMenuItem item;

        for (Plugin p : plugins)
            if (isActivePlugin(p.getName()))
                for (final ExternAction action : p.getActions()) {
                    item = new JMenuItem(action.getMessage());
                    item.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent e) {
                            action.run();
                        }
                    });
                    menu.add(item);
                }

        item = new JMenuItem("Configure plugins");
        item.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                final JDialog dialog = new JDialog(Application.getFrame());
                dialog.setTitle("Configure plugins");

                final JPanel panel = new JPanel();
                panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
                final JCheckBox[] check = new JCheckBox[plugins.size()];

                if (plugins.size() == 0) {
                    panel.add(new JLabel("No plugin found."));
                } else {
                    int i = 0;
                    for (Plugin p : plugins) {
                        check[i] = new JCheckBox(p.getName());
                        if (isActivePlugin(p.getName()))
                            check[i].setSelected(true);
                        else
                            check[i].setSelected(false);
                        panel.add(check[i]);
                        ++i;
                    }
                }

                JButton button = new JButton("OK");
                button.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        int i = 0;
                        for (Plugin p : plugins) {
                            setActivatedPlugin(p.getName(), check[i].isSelected());
                            Application.updatePluginsMenu();
                            ++i;
                        }
                        dialog.setVisible(false);
                    }
                });

                panel.add(button);

                JScrollPane scroll = new JScrollPane(panel);
                dialog.add(scroll);
                dialog.pack();
                dialog.setLocationRelativeTo(null);
                dialog.setVisible(true);
            }
        });

        menu.addSeparator();
        menu.add(item);

        return menu;
    }

}
