package project;

import java.util.*;
import java.io.FileNotFoundException;

import language.java.JavaCompiler;
import language.java.JavaRunner;

import application.*;
import static logging.Logging.logger;

public class ProjectManager {

    private static LinkedList <File> openedFiles = new LinkedList <File> ();
    private static LinkedList <Project> openedProjects = new LinkedList <Project> ();

    public static void openFile(String path) {
        // Check if the file is already opened.
        for (File f : openedFiles)
            if (path.equals(f.getPath())) {
                logger.info("File " + path + " is already opened.");

                Editor editor = Application.getEditor();
                for (int i = 0; i < editor.getTabCount(); i++) {
                    EditorPane editorPane = ((EditorContent) editor.getComponentAt(i)).getEditorPane();
                    if (path.equals(editorPane.getPath())) {
                        editor.setSelectedIndex(i);
                        editorPane.grabFocus();
                        break;
                    }
                }

                return;
           }

        // Load the file.
        File f;
        try {
            f = new File(path);
        } catch (java.io.IOException e) {
            logger.throwing("project.ProjectManager", "openFile", e);
            return;
        }

        // We have a file. Let's use it.
        openedFiles.add(f);
        f.addObserver(Application.addEditorTab(f.getName()));
        f.load();
        Config.addFile(path);
    }

    public static void openFile(String file, String project) {
        Project p = getProject(project);
        if (p != null)
            openFile(p.getPathToDir() + file);
    }

    public static void closeFile(String path) {
        for (File f : openedFiles)
            if (path.equals(f.getPath())) {
                openedFiles.remove(f);
                return;
            }
    }

    public static void saveFile(String path, String data) {
        try {
            File f = new File(path);
            f.save(data);
        } catch (java.io.IOException e) {
            logger.throwing("ProjectManager", "saveFile", e);
        }
    }

    public static void saveFiles(Iterable<String> files) {
        Editor editor = Application.getEditor();
        int nr = editor.getTabCount();
        for (String file : files)
            for (int i = 0; i < nr; i++) {
                EditorPane editorPane = ((EditorContent) editor.getComponentAt(i)).getEditorPane();
                if (file.equals(editorPane.getPath())) {
                    saveFile(file, editorPane.getText());
                    editorPane.setSaved();
                    break;
                }
            }
    }

    // The function with 2 parameters is for new projects only.
    public static void openProject(String path) {
        openProject(path, null);
    }

    public static void openProject(String path, String title) {
        // Check if the project is already opened.
        for (Project p : openedProjects)
            if (path.equals(p.getPath())) {
                logger.info("Project " + path + " is already opened.");
                return;
            }

        // Load the project.
        Project p;
        try {
            p = new Project(path, title);
        } catch (java.io.IOException e) {
            logger.throwing("project.ProjectManager", "openProject", e);
            return;
        }

        // We have a project. Let's use it.
        openedProjects.add(p);
        p.addObserver(Application.getProjectsTree());
        p.load();
    }

    public static void closeProject(String title) {
        Project p = getProject(title);
        if (p != null) {
            openedProjects.remove(p);
            Application.getProjectsTree().update(null, null);
        }
    }

    public static void addFileToProject(String source, String file, String project)
                                                                throws FileNotFoundException {
        if (source != null && !File.exists(source))
            throw new FileNotFoundException("Source file " + source + " does not exist.");

        Project p = getProject(project);
        if (p != null) {
            try {
                String path = p.getPathToDir() + file;
                (new File(path)).copy(source);
                p.addFile(file);
            } catch (java.io.IOException e) {
                logger.throwing("ProjectManager", "addFileToProject", e);
            }
        }
    }

    public static void addFileToProject(String file, String project) {
        try {
            addFileToProject(null, file, project);
        } catch (FileNotFoundException e) {
            // This will never happen
        }
    }

    public static void addFolderToProject(String folder, String project) {
        Project p = getProject(project);

        if (p != null) {
            String path = p.getPathToDir() + folder;
            (new java.io.File(path)).mkdir();
            p.addFolder(folder);
        }
    }

    public static void removeFileFromProject(String file, String project) {
        Project p = getProject(project);

        if (p != null) {
            File.delete(p.getPathToDir() + file);
            p.removeFile(file);
        }
    }

    public static void moveFile(String from, String to, String project) throws FileNotFoundException {
        Project p = getProject(project);

        if (p != null) {
            java.io.File f = (new java.io.File(p.getPathToDir() + from));
            if (!f.exists())
                throw new FileNotFoundException("The soure file " + f.getAbsolutePath() + " does not exist.");
            f.renameTo(new java.io.File(p.getPathToDir() + to));
            p.removeFile(from);
            p.addFile(to);
        }
    }

    public static boolean compileProject(Project p) {
        if (p != null) {
            File.recreate(p.getPathToDir() + "bin");
            Iterable<String> files = p.getFilePaths();
            saveFiles(files);
            return JavaCompiler.compile(files, p);
        }

        return false;
    }

    public static void compileProject(String project) {
        compileProject(getProject(project));
    }

    public static void compileProjectForFile(String filePath) {
        compileProject(getProjectForFile(filePath));
    }

    // runs project if it has only a Main class (returns false if the Main class is not found)
    public static boolean compileAndRunProjectForFile(String filePath, String args) {
        Project p = getProjectForFile(filePath);

        if (p != null) {
            if (!compileProject(p))
                return true;

            String mainClass = p.getMainClass();
            if (mainClass == null)
                return false;

            JavaRunner.run(p, mainClass.substring(0, mainClass.lastIndexOf('.')), args);

            return true;
        }
        return true;
    }

    // runs project if it has only a Main class (returns false if the Main class is not found)
    public static boolean compileAndRunProject(String project, String args) {
        Project p = getProject(project);

        if (p != null) {
            if (!compileProject(p))
                return true;

            String mainClass = p.getMainClass();

            if (mainClass == null)
                return false;

            JavaRunner.run(p, mainClass.substring(0, mainClass.lastIndexOf('.')), args);

            return true;
        }
        return true;
    }

    // runs project with custom main class
    public static void runProjectMain(String mainClass, String args) {
        Project p = getProjectForFile(mainClass);

        if (p != null) {
            JavaRunner.run(p, mainClass.substring(p.getPathToDir().length() + 4, mainClass.lastIndexOf('.')), args);
        }
    }

    public static boolean isDirectory(String path, String project) {
        Project p = getProject(project);

        if (p != null) {
            java.io.File f = (new java.io.File(p.getPathToDir() + path));
            return f.isDirectory();
        }
        return false;
    }

    public static LinkedList <ProjectStructure> getProjectTrees() {
        LinkedList <ProjectStructure> ret = new LinkedList <ProjectStructure> ();

        for (Project p : openedProjects)
            ret.add(p.getStructure());

        return ret;
    }

    public static Project getProject(String project) {
        for (Project p : openedProjects)
            if (project.equals(p.getTitle()))
                return p;
        logger.severe("Project " + project + " is not opened.");
        return null;
    }

    public static Project getProjectForFile(String filePath) {
        for (Project p : openedProjects) {
            String path = p.getPathToDir();
            if (filePath.indexOf(path) == 0)
                return p;
        }
        logger.severe("File " + filePath + " is not contained in any opened project.");
        return null;
    }

    public static LinkedList<String> getAllFilePaths(String filePath) {
        Project p = getProjectForFile(filePath);
        if (p != null)
            return p.getAllFilePaths();
        return new LinkedList<String>();
    }

    public static String getProjectPathForFile(String filePath) {
        for (Project p : openedProjects) {
            String path = p.getPathToDir();
            if (filePath.indexOf(path) == 0)
                return path + "src/";
        }
        logger.severe("File " + filePath + " is not contained in any opened project.");
        return null;
    }

    public static String getProjectBinPathForFile(String filePath) {
        for (Project p : openedProjects) {
            String path = p.getPathToDir();
            if (filePath.indexOf(path) == 0)
                return path + "bin/";
        }
        logger.severe("File " + filePath + " is not contained in any opened project.");
        return null;
    }

    public static String getProjectBinPath(String project) {
        Project p = getProject(project);
        if (p != null)
           return p.getPathToDir() + "bin/";
        return null;
    }

    public static EditorPane getFileEditor(String path) {
        openFile(path);
        int index = 0;
        Editor editor = Application.getEditor();
        while (index < editor.getTabCount()) {
            EditorContent editorContent = (EditorContent) editor.getComponentAt(index);
            if (editorContent != null) {
                EditorPane editorPane = editorContent.getEditorPane();
                if (editorPane.getPath().equals(path)) {
                    editor.setSelectedComponent(editorContent);
                    return editorPane;
                }
            }
            index++;
        }
        return null;
    }

}
